import numpy as np

from treesumstats import BASIC_METRIC2HELP, BASIC_ARRAY_METRICS, BASIC_METRIC2FUN, FeatureCalculator, NORMALIZED
from treesumstats.tree_manager import annotate_forest_with_time, TIME


class EventTimeFeatureCalculator(FeatureCalculator):
    """Computes tree event time-related summary statistics."""

    def __init__(self):
        self._max_time = None
        self._tip_times = None
        self._forest = None
        self._inode_times = None

    def feature_names(self):
        return [f'{TIME}_tip_{flavour}' for flavour in BASIC_ARRAY_METRICS] \
            +  [f'{TIME}_inode_{flavour}' for flavour in BASIC_ARRAY_METRICS] \
            +  [f'{TIME}_tip_{NORMALIZED}_{flavour}' for flavour in BASIC_ARRAY_METRICS] \
            + [f'{TIME}_inode_{NORMALIZED}_{flavour}' for flavour in BASIC_ARRAY_METRICS]

    def set_forest(self, forest, **kwargs):
        self._forest = forest
        annotate_forest_with_time(self._forest)
        self._tip_times = None
        self._inode_times = None
        self._max_time = None

    @property
    def tip_times(self):
        if self._tip_times is None:
            self._tip_times = np.array([getattr(t, TIME) for tree in self._forest for t in tree])
        return self._tip_times

    @property
    def max_time(self):
        if self._max_time is None:
            self._max_time = np.max(self.tip_times)
        return self._max_time

    @property
    def inode_times(self):
        if self._inode_times is None:
            self._inode_times \
                = np.array([getattr(n, TIME) for tree in self._forest for n in tree.traverse() if not n.is_leaf()])
        return self._inode_times

    def calculate(self, feature_name, **kwargs):
        if feature_name.startswith(TIME):
            is_internal = 'inode' in feature_name
            is_normalized = NORMALIZED in feature_name
            flavour = feature_name.split('_')[-1]
            return BASIC_METRIC2FUN[flavour]((self.inode_times if is_internal else self.tip_times)
                                             / (self.max_time if is_normalized else 1))

        return None

    def help(self, feature_name, *args, **kwargs):
        if feature_name.startswith(TIME):
            is_internal = 'inode' in feature_name
            node = 'internal node' if is_internal else 'tip'
            is_normalized = NORMALIZED in feature_name
            flavour = feature_name.split('_')[-1]
            res = f'{BASIC_METRIC2HELP[flavour]} of the sampling times of forest {node}s.'
            if is_normalized:
                res += ' We normalize the sampling times by dividing them by the last tip\'s sampling time.'
            return res
        return None


from collections import Counter

import numpy as np
from scipy.special import digamma

from treesumstats import FeatureCalculator, NORMALIZED

IMBALANCE_AVG = 'imbalance_avg'

DEPTH_MAX = 'depth_max'

WIDTH_DELTA = 'width_delta'

WIDTH_DEPTH_RATIO = 'width_depth_ratio'

WIDTH_MAX = 'width_max'

SACKIN = 'sackin'

COLLESS = 'colless'

DEPTH = "depth"

class TopologyFeatureCalculator(FeatureCalculator):
    """Computes tree topology-related summary statistics."""

    def __init__(self):
        self._in_ladder = None
        self._max_ladder_norm = None
        self._max_ladder = None
        self._n_inodes = None
        self._max_width = None
        self._max_depth = None
        self._sackin = None
        self._colless = None
        self._forest = None
        self._n_tips = None
        self._delta_w = None
        self._avg_imbalance = None
        self._n_imbalanced = None

    def help(self, feature_name, *args, **kwargs):
        if feature_name.startswith(COLLESS):
            res = f'sum (over internal nodes) of absolute differences of numbers of tips in their left and right subtrees [Colless 1982].'
            if feature_name.endswith(NORMALIZED):
                res += ' We normalize it by dividing it by this metric for a forest of a ladder and all the other trees only having 1 tip: '\
                       '(n_tips - n_trees - 1) (n_tips - n_trees) / 2.'
            return res
        if feature_name.startswith(SACKIN):
            res = f'sum (over tips) of the number of internal nodes separating them from the root [Sackin 1972].'
            if feature_name.endswith(NORMALIZED):
                res += ' We normalize it by dividing it the value of this metric for a forest of a ladder and all the other trees only having 1 tip: '\
                       '(n_tips - n_trees) (n_tips - n_trees + 1) / 2 - 1.'
            return res

        if feature_name.startswith(DEPTH_MAX):
            res = "maximum node depth (i.e., number of branches separating a node from its tree root) [Colijn et al. 2014]."
            if feature_name.endswith(NORMALIZED):
                res += ' We normalize it by dividing it by the value of this metric for a forest of a ladder and all the other trees only having 1 tip: '\
                       '(n_tips - n_trees).'
            return res
        if feature_name.startswith(WIDTH_MAX):
            res = 'maximum width, where the width is defined as the number of nodes that are at the same depth, '\
                  'and node depth as the number of branches separating a node from its tree root [Colijn et al. 2014].'
            if feature_name.endswith(NORMALIZED):
                res += ' We normalize it by dividing it by the value of this metric for a forest of the same number of balanced trees with the same total number of tips.'
            return res
        if feature_name.startswith(WIDTH_DEPTH_RATIO):
            res = 'maximum width to maximum depth ratio, where the width is defined as the number of nodes that are at the same depth, '\
                  'and node depth as the number of branches separating a node from its tree root [Colijn et al. 2014].'
            if feature_name.endswith(NORMALIZED):
                res += ' We normalize it by dividing it by the value of this metric for a forest of the same number of balanced trees with the same total number of tips.'
            return res
        if feature_name.startswith(WIDTH_DELTA):
            res = 'maximum width difference between neighbouring depths, ' \
                  'where the width is defined as the number of nodes that are at the same depth, ' \
                  'and node depth as the number of branches separating a node from its tree root [Colijn et al. 2014].'
            if feature_name.endswith(NORMALIZED):
                res += (' We normalize it by dividing it by the value of this metric for a forest of the same '
                        'number of balanced trees and last two layers reorganised to increase the delta (tips at the last layer are grouped into one ladderized subtree).')
            return res
        if 'ladder' in feature_name:
            if 'inodes_in' in feature_name:
                measure = 'fraction' if feature_name.startswith('frac') else 'number'
                return f'{measure} of forest internal nodes that are resolved and have a tip descendant [Colijn et al. 2014].'
            elif 'max' in feature_name:
                res = 'maximum number of connected internal nodes with a single leaf descendant [Colijn et al. 2014].'
                if feature_name.endswith(NORMALIZED):
                    res += ' We calculate it separately on each forest\'s tree and divide it by the number of leaves in that tree. '\
                           'The maximum of of these values is then reported.'
                return res
        if 'imbalanced' in feature_name:
            measure = 'fraction' if feature_name.startswith('frac') else 'number'
            return f'{measure} of internal nodes that have different number of leaves ' \
                   f'in their smallest and their largest child subtrees [Colijn et al. 2014].'
        if IMBALANCE_AVG == feature_name:
            return 'mean ratio of min-to-max subtree sizes over all internal nodes [Colijn et al. 2014].'
        return None

    def feature_names(self):
        return [COLLESS, f'{COLLESS}_{NORMALIZED}'] \
            + [SACKIN, f'{SACKIN}_{NORMALIZED}'] \
            + [WIDTH_MAX, f'{WIDTH_MAX}_{NORMALIZED}'] \
            + [DEPTH_MAX, f'{DEPTH_MAX}_{NORMALIZED}'] \
            + [WIDTH_DEPTH_RATIO, f'{WIDTH_DEPTH_RATIO}_{NORMALIZED}'] \
            + [WIDTH_DELTA, f'{WIDTH_DELTA}_{NORMALIZED}'] \
            + ['frac_inodes_in_ladder', 'n_inodes_in_ladder',
               'len_ladder_max', f'len_ladder_max_{NORMALIZED}'] \
            + ['frac_inodes_imbalanced', 'n_inodes_imbalanced'] \
            + [IMBALANCE_AVG]

    def set_forest(self, forest, **kwargs):
        self._forest = forest
        self._n_tips = None
        self._colless = None
        self._sackin = None
        self._max_depth = None
        self._max_width = None
        self._delta_w = None
        self._n_inodes = None
        self._max_ladder_norm = None
        self._max_ladder = None
        self._in_ladder = None
        self._avg_imbalance = None
        self._n_imbalanced = None

    @property
    def n_tips(self):
        if self._n_tips is None:
            self._n_tips = sum(len(tree) for tree in self._forest)
        return self._n_tips

    @property
    def n_inodes(self):
        """Number of internal nodes in the forest"""
        if self._n_inodes is None:
            self._n_inodes = sum(sum(1 for _ in tree.traverse() if not _.is_leaf()) for tree in self._forest)
        return self._n_inodes


    @property
    def colless(self):
        if self._colless is None:
            self._colless = 0

            for tree in self._forest:
                for node in tree.traverse():
                    if not node.is_leaf():
                        if len(node.children) == 2:
                            child1, child2 = node.children
                            self._colless += abs(len(child1) - len(child2))
                        else:
                            scores = []
                            for j in range(len(node.children)):
                                for k in range(j + 1, len(node.children)):
                                    scores.append(abs(len(node.children[j]) - len(node.children[k])))
                            self._colless += np.average(scores)
        return self._colless

    def _annotate_depth(self):
        if not hasattr(self._forest[0], DEPTH):
            for tree in self._forest:
                for node in tree.traverse('preorder'):
                    parent_depth = -1 if node.is_root() else getattr(node.up, DEPTH)
                    node.add_feature(DEPTH, parent_depth + 1)

    @property
    def sackin(self):
        if self._sackin is None:
            self._annotate_depth()
            self._sackin = sum(sum(getattr(tip, DEPTH) - 1 for tip in tree if not tip.is_root())
                               for tree in self._forest)
        return self._sackin

    @property
    def max_depth(self):
        if self._max_depth is None:
            self._annotate_depth()
            self._max_depth = max(max(getattr(tip, DEPTH) for tip in tree) for tree in self._forest)
        return self._max_depth

    def _set_depth_delta(self):
        self._annotate_depth()
        width_d = Counter(getattr(node, DEPTH) for tree in self._forest for node in tree.traverse())
        self._max_width = width_d.most_common(n=1)[0][1]

        if self._max_depth is None:
            self._max_depth = max(width_d.keys())

        self._delta_w = 0
        for d in range(1, self._max_depth + 1):
            self._delta_w = max(np.abs(width_d[d] - width_d[d - 1]), self._delta_w)

    @property
    def max_width(self):
        if self._max_width is None:
            self._set_depth_delta()
        return self._max_width

    @property
    def delta_w(self):
        if self._delta_w is None:
            self._set_depth_delta()
        return self._delta_w

    def max_width_depth_balanced(self):
        # we distribute leaves equally into n_forest balanced trees
        # and compare the number of nodes at the second-but-last and the last levels to calculate the max width
        n_trees = len(self._forest)
        n_nodes = 2 * self.n_tips - n_trees
        l = np.floor(np.log2(self.n_tips / n_trees))
        nodes_in_second_but_last_level = n_trees * np.power(2, l)
        nodes_in_all_but_last_levels = n_trees * (np.power(2, l + 1) - 1)
        nodes_in_last_level = n_nodes - nodes_in_all_but_last_levels
        mw = max(nodes_in_second_but_last_level, nodes_in_last_level)
        md = l + 1 if nodes_in_last_level else l
        # to calculate the max_delta we take the level with the max_width
        # and (if it was not the last one) ladderize the leaves that belonged to the last level (hence leaving only two leaves there);
        # or (otherwise) removing a cherry from the last level and regrafting it on top of one of its existing tips
        # (hence removing two leaves from that level and adding two to the next one)
        if n_nodes == n_trees:
            delta = 0
        else:
            if not nodes_in_last_level:
                delta = max(nodes_in_second_but_last_level - 4,
                            nodes_in_second_but_last_level - nodes_in_second_but_last_level / 2)
            else:
                delta = max(np.abs(nodes_in_last_level - nodes_in_second_but_last_level),
                            nodes_in_second_but_last_level - 2,
                            nodes_in_last_level - 4)
        return mw, md, delta

    @property
    def in_ladder(self):
        if self._in_ladder is None:
            self._set_ladder()
        return self._in_ladder

    @property
    def max_ladder(self):
        if self._max_ladder is None:
            self._set_ladder()
        return self._max_ladder

    @property
    def max_ladder_norm(self):
        if self._max_ladder_norm is None:
            self._set_ladder()
        return self._max_ladder_norm

    def _set_imbalance(self):
        self._n_imbalanced = 0
        self._avg_imbalance = 0

        for tree in self._forest:
            for node in tree.traverse():
                if not node.is_leaf():
                    c_lens = sorted([len(c) for c in node.children])
                    if c_lens[0] != c_lens[-1]:
                        self._n_imbalanced += 1
                    self._avg_imbalance += c_lens[0] / c_lens[-1]
        self._avg_imbalance /= self.n_inodes

    @property
    def n_imbalanced(self):
        if self._n_imbalanced is None:
            self._set_imbalance()
        return self._n_imbalanced

    @property
    def avg_imbalanced(self):
        if self._avg_imbalance is None:
            self._set_imbalance()
        return self._avg_imbalance


    def _set_ladder(self):
        self._max_ladder = 0
        self._max_ladder_norm = 0
        self._in_ladder = 0
        for tree in self._forest:
            cur_ladder = 0
            for node in tree.traverse('preorder'):
                if 2 == len(node.children) and next((True for c in node.children if c.is_leaf()), False):
                    self._in_ladder += 1
                    parent_ladder_len = 0 if node.is_root() else getattr(node.up, 'ladder', 0)
                    ladder_len = 1 + parent_ladder_len
                    node.add_feature('ladder', ladder_len)
                    cur_ladder = max(cur_ladder, ladder_len)
            for node in set(_.up for _ in tree if not _.is_root()):
                node.del_feature('ladder')
            self._max_ladder = max(cur_ladder, self._max_ladder)
            self._max_ladder_norm = max(cur_ladder / len(tree), self._max_ladder_norm)



    def calculate(self, feature_name, **kwargs):
        n_trees = len(self._forest)
        if feature_name.startswith(COLLESS):
            denominator = (self.n_tips - n_trees - 1) * (self.n_tips - n_trees) / 2 \
                if feature_name.endswith(NORMALIZED) else 1
            return self.colless / denominator
        if feature_name.startswith(SACKIN):
            denominator = (self.n_tips - n_trees) * (self.n_tips - n_trees + 1) / 2 - 1 \
                if feature_name.endswith(NORMALIZED) else 1
            return self.sackin / denominator
        if feature_name.startswith(DEPTH_MAX):
            denominator = (self.n_tips - n_trees) if feature_name.endswith(NORMALIZED) else 1
            return self.max_depth / denominator
        if feature_name.startswith(WIDTH_MAX):
            denominator = self.max_width_depth_balanced()[0] if feature_name.endswith(NORMALIZED) else 1
            return self.max_width / denominator
        if feature_name.startswith(WIDTH_DEPTH_RATIO):
            mw, md, _ = self.max_width_depth_balanced()
            denominator = mw / md if feature_name.endswith(NORMALIZED) else 1
            return self.max_width / self.max_depth / denominator
        if feature_name.startswith(WIDTH_DELTA):
            denominator = self.max_width_depth_balanced()[-1] if feature_name.endswith(NORMALIZED) else 1
            return self.delta_w / denominator
        if 'ladder' in feature_name:
            if 'inodes_in' in feature_name:
                denominator = self.n_inodes if feature_name.startswith('frac') else 1
                return self.in_ladder / denominator
            elif 'max' in feature_name:
                return self.max_ladder_norm if feature_name.endswith(NORMALIZED) else self.max_ladder
        if 'imbalanced' in feature_name:
            denominator = self.n_inodes if feature_name.startswith('frac') else 1
            return self.n_imbalanced / denominator
        if IMBALANCE_AVG == feature_name:
            return self.avg_imbalanced
        return None


def min_avg_stair(n):
    return digamma(n + 1) + np.euler_gamma


import numpy as np

from treesumstats import BASIC_METRIC2HELP, BASIC_ARRAY_METRICS, BASIC_METRIC2FUN, FeatureCalculator

FRACTION_RESOLVED_NON_ZERO = "frac_inodes_resolved_non_zero"

FRACTION_RESOLVED = "frac_inodes_resolved"

N_RESOLVED_NON_ZERO = "n_inodes_resolved_non_zero"

N_RESOLVED = "n_inodes_resolved"

N_CHILDREN_PREFIX = "n_children"


class ResolutionFeatureCalculator(FeatureCalculator):
    """Computes tree resolution-related summary statistics."""

    def __init__(self):
        self._n_resolved_max = None
        self._n_resolved_nz = None
        self._n_resolved = None
        self._forest = None
        self._n_children = None

    def feature_names(self):
        return ([f'{N_CHILDREN_PREFIX}_{flavour}' for flavour in BASIC_ARRAY_METRICS] +
                [N_RESOLVED, N_RESOLVED_NON_ZERO, FRACTION_RESOLVED, FRACTION_RESOLVED_NON_ZERO])

    def set_forest(self, forest, **kwargs):
        self._forest = forest
        self._n_children = None
        self._n_resolved = None
        self._n_resolved_nz = None
        self._n_resolved_max = None

    @property
    def n_children(self):
        if self._n_children is None:
            n_children = []
            for tree in self._forest:
                n_children.extend(len(n.children) for n in tree.traverse() if not n.is_leaf())
            self._n_children =  np.array(n_children, dtype=int)
        return self._n_children

    @property
    def n_resolved(self):
        if self._n_resolved is None:
            self._n_resolved = 0
            self._n_resolved_nz = 0
            for tree in self._forest:
                for n in tree.traverse():
                    if 2 == len(n.children):
                        self._n_resolved += 1
                        if n.is_root() or n.dist:
                            if all(_.dist > 0 for _ in n.children):
                                self._n_resolved_nz += 1
        return self._n_resolved

    @property
    def n_resolved_binary(self):
        if self._n_resolved_max is None:
            self._n_resolved_max = sum(len(tree) for tree in self._forest) - len(self._forest)
        return self._n_resolved_max

    @property
    def n_resolved_nz(self):
        assert self.n_resolved >= 0
        return self._n_resolved_nz

    def calculate(self, feature_name, **kwargs):
        if feature_name.startswith(N_CHILDREN_PREFIX):
            return BASIC_METRIC2FUN[feature_name[len(N_CHILDREN_PREFIX) + 1:]](self.n_children)
        if N_RESOLVED == feature_name:
            return self.n_resolved
        if FRACTION_RESOLVED == feature_name:
            return self.n_resolved / self.n_resolved_binary
        if N_RESOLVED_NON_ZERO == feature_name:
            return self.n_resolved_nz
        return self.n_resolved_nz / self.n_resolved_binary

    def help(self, feature_name, *args, **kwargs):
        if feature_name.startswith(N_CHILDREN_PREFIX):
            return (f'{BASIC_METRIC2HELP[feature_name[len(N_CHILDREN_PREFIX) + 1:]]} '
                    f'of the number of children per internal node of forest trees.')
        if N_RESOLVED == feature_name:
            return 'number of resolved (i.e., with exactly 2 children) nodes in the forest.'
        if N_RESOLVED_NON_ZERO == feature_name:
            return ('number of resolved (i.e., with exactly 2 children) forest nodes '
                    'that are either root nodes or have non-zero branch length.')
        if FRACTION_RESOLVED == feature_name:
            return ('fraction of resolved (i.e., with exactly 2 children) nodes in the forest '
                    'with respect to the number of internal nodes in this forest '
                    'if all its trees were fully resolved (i.e., binary).')
        if FRACTION_RESOLVED_NON_ZERO == feature_name:
            return ('fraction of resolved (i.e., with exactly 2 children) forest nodes '
                    'that are either root nodes or have non-zero branch length '
                    'with respect to the number of internal nodes in this forest '
                    'if all its trees were fully resolved (i.e., binary).')
        return None



from treesumstats import FeatureCalculator


class BasicFeatureCalculator(FeatureCalculator):
    """Computes tree height-related summary statistics."""

    def __init__(self):
        self._forest = None

    def feature_names(self):
        return ['n_trees', 'n_tips', 'n_inodes', 'len_forest']

    def set_forest(self, forest, **kwargs):
        self._forest = forest

    def calculate(self, feature_name, **kwargs):
        if 'n_trees' == feature_name:
            return len(self._forest)
        if 'n_tips' == feature_name:
            return sum(len(tree) for tree in self._forest)
        if 'n_inodes' == feature_name:
            return sum(1 for tree in self._forest for n in tree.traverse() if not n.is_leaf())
        if 'len_forest' == feature_name:
            return sum(_.dist for tree in self._forest for _ in tree.traverse())
        return None

    def help(self, feature_name, *args, **kwargs):
        if 'n_trees' == feature_name:
            return 'number of trees in the forest.'
        if 'n_tips' == feature_name:
            return 'number of tips in the forest.'
        if 'n_inodes' == feature_name:
            return 'number of internal nodes in the forest.'
        if 'len_forest' == feature_name:
            return 'sum of forest branch lengths.'
        return None



import warnings
from abc import ABC, abstractmethod

import numpy as np

MED = 'median'

VAR = 'var'

MEAN = 'mean'

MAX = 'max'

MIN = 'min'

EPSILON = 1e-6


BASIC_ARRAY_METRICS = [MEAN, MIN, MAX, VAR, MED]

BASIC_METRIC2HELP = {VAR: 'variance',
                     MIN: 'minimum',
                     MEAN: 'average',
                     MED: 'median',
                     MAX: 'maximum'}

BASIC_METRIC2FUN = {VAR: np.var,
                    MIN: np.min,
                    MEAN: np.mean,
                    MED: np.median,
                    MAX: np.max}


NORMALIZED = 'normalized'


class FeatureCalculator(ABC):
    """Abstract base class for feature calculators."""

    @abstractmethod
    def feature_names(self):
        """Return a list of feature names this class can compute."""
        pass

    @abstractmethod
    def set_forest(self, forest, **kwargs):
        """Sets the forest on which the features are to be calculated"""
        pass

    @abstractmethod
    def calculate(self, forest, feature_name, **kwargs):
        """Compute a specific feature or features."""
        pass

    @abstractmethod
    def help(self, feature_name, *args, **kwargs):
        """Shows the definition of the specific feature."""
        pass


class FeatureRegistry:
    """Registry for mapping feature names to their respective calculator classes."""

    _registry = {}
    _calculators = []

    @classmethod
    def register(cls, calculator):
        """Register a feature calculator class and map its features."""
        cls._calculators.append(calculator)
        for feature in calculator.feature_names():
            cls._registry[feature] = calculator

    @classmethod
    def get_calculator(cls, feature_name):
        """Retrieve the appropriate calculator class for a given feature."""
        return cls._registry.get(feature_name)

    @classmethod
    def list_features(cls):
        return list(cls._registry.keys())

    @classmethod
    def list_calculators(cls):
        return list(cls._calculators)


class FeatureManager:
    """Manages feature computation dynamically based on user input."""

    @classmethod
    def compute_features(cls, forest, *feature_list, **kwargs):
        """Compute and return requested features."""

        for calc in FeatureRegistry.list_calculators():
            calc.set_forest(forest, **kwargs)

        if not feature_list:
            feature_list = FeatureManager.available_features()

        for feature in feature_list:
            calculator = FeatureRegistry.get_calculator(feature)
            if calculator:
                yield calculator.calculate(feature, **kwargs)
            else:
                warnings.warn(f'Did not find a Feature Calculator for feature {feature}')
                yield None

    @classmethod
    def available_features(cls):
        return FeatureRegistry.list_features()

    @classmethod
    def help(cls, feature_name, *args, **kwargs):
        """Returns feature description"""

        calculator = FeatureRegistry.get_calculator(feature_name)
        if calculator:
            return calculator.help(feature_name, **kwargs)
        else:
            warnings.warn(f'Did not find a Feature Calculator for feature {feature_name}')
            return None






import re
from collections import defaultdict

import numpy as np
import scipy

from treesumstats import FeatureCalculator, BASIC_METRIC2FUN, BASIC_METRIC2HELP, BASIC_ARRAY_METRICS
from treesumstats.tree_manager import annotate_forest_with_time, TIME

OTHER_CONFIGURATION_LABEL = 'O'

TOPOLOGY_LABELS = ['2', '3U', '3L', '4U', '4U3U1', '4U211', '4L', '4B', 'O']

TIP = 1

CHERRY = (TIP, TIP)

TRIPLET_UNRESOLVED = (TIP, TIP, TIP)
TRIPLET_LADDER = (CHERRY, TIP)

QUARTET_UNRESOLVED = (TIP, TIP, TIP, TIP)
QUARTET_TRIPLET_UNRESOLVED_TIP = (TRIPLET_UNRESOLVED, TIP)
QUARTET_CHERRY_TIP_TIP = (CHERRY, TIP, TIP)
QUARTET_LADDER = (TRIPLET_LADDER, TIP)
QUARTET_BALANCED = (CHERRY, CHERRY)

OTHER = 2

CONFIG_ORDER = [TIP, CHERRY, TRIPLET_UNRESOLVED, TRIPLET_LADDER,
                QUARTET_UNRESOLVED, QUARTET_TRIPLET_UNRESOLVED_TIP, QUARTET_CHERRY_TIP_TIP, QUARTET_LADDER, QUARTET_BALANCED,
                OTHER]

TOPOLOGY_LABEL2KEY = {'2': CHERRY, '3U': TRIPLET_UNRESOLVED, '3L': TRIPLET_LADDER, '4U': QUARTET_UNRESOLVED,
                      '4U3U1': QUARTET_TRIPLET_UNRESOLVED_TIP, '4U211': QUARTET_CHERRY_TIP_TIP,
                      '4L': QUARTET_LADDER, '4B': QUARTET_BALANCED, 'O': 1}

TOPOLOGY_LABEL2N = {'2': 2, '3U': 3, '3L': 3, '4U': 4,
                    '4U3U1': 4, '4U211': 4,
                    '4L': 4, '4B': 4, 'O': 1}

TOPOLOGY_LABEL2NAME = {'2': 'cherries', '3U': 'unresolved triplets', '3L': 'resolved triplets',
                       '4U': 'fully unresolved quartets',
                       '4U3U1': 'partially resolved quartets with an unresolved triplet and a tip as children',
                       '4U211': 'partially resolved quartets with a cherry and two tips as three children',
                       '4L': 'ladderized fully resolved quartets',
                       '4B': 'balanced quartets (with two cherries as children)'}


TIP_TIME_PERCENTILES = [1, 5, 10, 25, 75, 90, 95, 99]


def get_config2tips(forest):
    topo2tips = defaultdict(list)
    feature = 'configuration'
    for tree in forest:
        for n in tree.traverse('postorder'):
            if n.is_leaf():
                if not n.is_root():
                    n.add_feature(feature, TIP)
                continue

            if len(n) > 4:
                topo2tips[OTHER].extend(_ for _ in n.children if _.is_leaf())
            else:
                config = tuple(
                    sorted([getattr(c, feature) for c in n.children], key=lambda _: -CONFIG_ORDER.index(_)))
                n.add_feature(feature, config)
                topo2tips[config].append(list(n.iter_leaves()))

            for c in n.children:
                c.del_feature(feature)
    return topo2tips

def get_reshuffled_configs(real_configs, all_tips_sorted):
    """
    Creates reshuffled configs. For each real config a reshuffled one is constructed by
        (1) keeping one tip from the real config
        (2) for each other tip from the real config picking a tip such that
            (2a) the time at the beginning of its branch is as close as possible to the time at the beginning of the real tip's branch,
            (2b) all the tips in the reshuffled config are different
            and (2c) there are no two tips in the reshuffled config that used to belong to the sam real config
                    (unless there are not enough tips, e.g., 8 tips in total and 2 4-tip configs)

    :param real_configs: an array of tip arrays representing real configs
    :param all_tips_sorted: an array of all tips in the forest, sorted by the time at the beginning of their branches (ascending order)
    :return: an array of tip arrays representing reshuffled configs
    """
    if not len(real_configs):
        return real_configs

    config_len = len(real_configs[0])
    tip2config = {t: i for i, conf in enumerate(real_configs) for t in conf}

    def get_start_time(tip):
        return getattr(tip, TIME) - tip.dist

    n_tips = len(all_tips_sorted)

    def get_closest_tip(tip, is_ok=lambda _: True):
        j = all_tips_sorted.index(tip)
        tip_stime = get_start_time(tip)

        next_tip = next((all_tips_sorted[_] for _ in range(j + 1, n_tips) if is_ok(all_tips_sorted[_])), None)
        prev_tip = next((all_tips_sorted[_] for _ in range(j - 1, -1, -1) if is_ok(all_tips_sorted[_])), None)
        return min([_ for _ in (next_tip, prev_tip) if not _ is None], key=lambda t: np.abs(get_start_time(t) - tip_stime))

    reshuffled_confs = []
    for i, conf in enumerate(real_configs):
        fixed_tip = np.random.choice(conf, size=1, replace=False)[0]
        used_tips = []
        used_configs = {i}

        def is_ok(tip):
            return (tip not in used_tips) \
                and (tip not in tip2config or tip2config[tip] not in used_configs \
                     or (len(used_configs) * config_len + (len(used_tips) - len(used_configs)) == n_tips))

        for other_tip in conf:
            tip = other_tip if other_tip == fixed_tip else get_closest_tip(other_tip, is_ok)
            used_tips.append(tip)
            if tip in tip2config:
                used_configs.add(tip2config[tip])
        reshuffled_confs.append(used_tips)

    return reshuffled_confs


def get_avg_time_diff(times):
    n = len(times)
    diff_sum = 0
    for i in range(n - 1):
        time_i = times[i]
        for j in range(i + 1, n):
            diff_sum += np.abs(time_i - times[j])
    return 2 * diff_sum / n / (n - 1)




class SubtreeFeatureCalculator(FeatureCalculator):
    """Computes tree topology-related summary statistics."""

    def __init__(self):
        self._topo2diffs = None
        self._topo2diffs_random = None
        self._all_tips = None
        self._forest = None
        self._topo2tips = None
        self._n_tips = None
        self._topo2tips_random = None

    def help(self, feature_name, *args, **kwargs):
        if 'tips_in_' in feature_name:
            conf = feature_name[feature_name.find('tips_in_') + len('tips_in_'):]
            measure = 'fraction' if feature_name.startswith('frac') else 'number'
            if 'O' == conf:
                return f'{measure} of forest tips that are not part of any subtree of size 2, 3 or 4.'
            return f'{measure} of forest tips that are found in {TOPOLOGY_LABEL2NAME[conf]}.'

        if feature_name.startswith('time_diff_in_'):
            conf = feature_name[feature_name.find('time_diff_in_') + len('time_diff_in_'):].split('_')[0]
            is_real='real' in feature_name

            reshuffling_expl = '(To construct a reshuffled subtree of a certain kind from a real one we keep one of its (randomly chosen) tips, ' \
                               'while for each other tip from the real subtree we replace it with a tip, for which:' \
                               '(i) the time at the beginning of its branch is as close as possible to the time at the beginning of the real tip\'s branch, ' \
                               '(ii) while all the tips in the reshuffled subtree are different ' \
                               'and (iii) there are no two tips in the reshuffled subtree that used to belong to the same real subtree of this kind ' \
                               '(unless there were not enough tips, e.g., 8 tips in total and only 2 4-tip subtrees in the forest).'

            flavour = feature_name.split('_')[-1]
            if flavour in BASIC_METRIC2FUN or 'perc' in flavour:
                res = f'{BASIC_METRIC2HELP[flavour]}' if flavour in BASIC_METRIC2FUN \
                    else f'{re.findall(r'\d+', flavour)[0]}th percentile'
                res += f' of average of all tip pair sampling time differences in a {TOPOLOGY_LABEL2NAME[conf]} among all {'' if is_real else 'reshuffled'}' \
                       f'{TOPOLOGY_LABEL2NAME[conf]} subtrees in the forest.'
                if not is_real:
                    res += reshuffling_expl
                return res
            else:
                if 'random_vs_real_n_less' in feature_name:
                    return f'number of {TOPOLOGY_LABEL2NAME[conf]}, ' \
                           'for which their average tip sampling times are smaller in real subtrees than ' \
                           f'in the corresponding reshuffled ones. {reshuffling_expl}'
                elif 'random_vs_real_frac_less' in feature_name:
                    return f'fraction of {TOPOLOGY_LABEL2NAME[conf]}, ' \
                           'for which their average tip sampling times are smaller in real subtrees than ' \
                           f'in the corresponding reshuffled ones. {reshuffling_expl}'
                elif 'random_vs_real_pval_less' in feature_name:
                    return f'p-value of the sign test checking that for {TOPOLOGY_LABEL2NAME[conf]}, ' \
                           'their average tip sampling times are indistinguishable in real subtrees and ' \
                           f'in the corresponding reshuffled ones ' \
                           f'(the alternative hypothesis is that some real ones are smaller). {reshuffling_expl}'

        return None

    def feature_names(self):
        return [f'frac_tips_in_{conf}' for conf in TOPOLOGY_LABELS] \
            + [f'n_tips_in_{conf}' for conf in TOPOLOGY_LABELS] \
            + [f'time_diff_in_{conf}_{type}_{flavour}' for type in ('real', 'random') for flavour in BASIC_ARRAY_METRICS for conf in TOPOLOGY_LABELS if conf != OTHER_CONFIGURATION_LABEL] \
            + [f'time_diff_in_{conf}_{type}_perc{p}' for type in ('real', 'random') for p in TIP_TIME_PERCENTILES for conf in TOPOLOGY_LABELS if conf != OTHER_CONFIGURATION_LABEL] \
            + [f'time_diff_in_{conf}_random_vs_real_{metric}_less' for metric in ('n', 'frac', 'pval') for conf in TOPOLOGY_LABELS if conf != OTHER_CONFIGURATION_LABEL]

    def set_forest(self, forest, **kwargs):
        self._forest = forest
        annotate_forest_with_time(self._forest)
        self._topo2tips = None
        self._n_tips = None
        self._all_tips = None
        self._topo2tips_random = None
        self._topo2diffs = None
        self._topo2diffs_random = None

    @property
    def n_tips(self):
        if self._n_tips is None:
            self._n_tips = sum(len(tree) for tree in self._forest)
        return self._n_tips

    def conf2tips(self, config_type, is_real=True):
        if is_real:
            if self._topo2tips is None:
                self._topo2tips = get_config2tips(self._forest)
            return self._topo2tips[config_type]

        if self._topo2tips_random is None:
            self._topo2tips_random = dict()
        if config_type not in self._topo2tips_random:
            self._topo2tips_random[config_type] = get_reshuffled_configs(self.conf2tips(config_type, is_real=True),
                                                                         self.all_tips)
        return self._topo2tips_random[config_type]

    def conf2time_diffs(self, config_type, is_real=True):
        if is_real:
            if self._topo2diffs is None:
                self._topo2diffs = {}
            if config_type not in self._topo2diffs:
                confs_real = self.conf2tips(config_type, is_real=True)
                self._topo2diffs[config_type] = np.array([get_avg_time_diff([getattr(_, TIME) for _ in c]) for c in confs_real])
            return self._topo2diffs[config_type]
        if self._topo2diffs_random is None:
            self._topo2diffs_random = {}
        if config_type not in self._topo2diffs_random:
            confs_real = self.conf2tips(config_type, is_real=True)
            confs_random = self.conf2tips(config_type, is_real=False)
            self._topo2diffs_random[config_type] = np.array([get_avg_time_diff([getattr(t, TIME) - t.dist + tr.dist
                                                                                for t, tr in zip(c, cr)])
                                                             for c, cr in zip(confs_real, confs_random)])
        return self._topo2diffs_random[config_type]

    @property
    def all_tips(self):
        if self._all_tips is None:
            self._all_tips = sorted([t for tree in self._forest for t in tree], key=lambda _: getattr(_, TIME) - _.dist)
        return self._all_tips

    def calculate(self, feature_name, **kwargs):
        if 'tips_in_' in feature_name:
            conf = feature_name[feature_name.find('tips_in_') + len('tips_in_'):]
            denominator = self.n_tips if feature_name.startswith('frac') else 1
            return len(self.conf2tips(TOPOLOGY_LABEL2KEY[conf], is_real=True)) * TOPOLOGY_LABEL2N[conf] / denominator

        if feature_name.startswith('time_diff_in_'):
            conf = feature_name[feature_name.find('time_diff_in_') + len('time_diff_in_'):].split('_')[0]
            time_diffs = self.conf2time_diffs(TOPOLOGY_LABEL2KEY[conf], is_real='real' in feature_name)


            flavour = feature_name.split('_')[-1]
            n = len(time_diffs)
            if flavour in BASIC_METRIC2FUN:
                return BASIC_METRIC2FUN[flavour](time_diffs) if n else 0
            elif 'perc' in flavour:
                perc = int(re.findall(r'\d+', flavour)[0])
                return np.percentile(time_diffs, perc) if n else 0
            else:
                time_diffs_random = self.conf2time_diffs(TOPOLOGY_LABEL2KEY[conf], is_real=False)
                n_random_less_than_real = (time_diffs_random < time_diffs).sum()
                if 'random_vs_real_n_less' in feature_name:
                    return n_random_less_than_real
                elif 'random_vs_real_frac_less' in feature_name:
                    return n_random_less_than_real / n if n else 0
                elif 'random_vs_real_pval_less' in feature_name:
                    return scipy.stats.binomtest(n_random_less_than_real, n=n, p=0.5,
                                                 alternative='less').pvalue if n else 1




from treesumstats import BASIC_METRIC2HELP, BASIC_METRIC2FUN, FeatureCalculator, MEAN, MED, VAR, EPSILON
from treesumstats.tree_manager import annotate_forest_with_time, TIME

BRLEN = 'brlen'

class BranchFeatureCalculator(FeatureCalculator):
    """Computes tree branch-related summary statistics."""

    def __init__(self):
        self._edists_top = None
        self._idists_top = None
        self._edists_bottom = None
        self._idists_bottom = None
        self._edists_middle = None
        self._idists_middle = None
        self._max_time = None
        self._idists = None
        self._edists = None
        self._forest = None

    def feature_names(self):
        return [f'{BRLEN}_inode_{flavour}' for flavour in [MEAN, MED, VAR]] \
            + [f'{BRLEN}_tip_{flavour}' for flavour in [MEAN, MED, VAR]] \
            + [f'{BRLEN}_inode_top_{flavour}' for flavour in [MEAN, MED, VAR]] \
            + [f'{BRLEN}_tip_top_{flavour}' for flavour in[MEAN, MED, VAR]]\
            + [f'{BRLEN}_inode_middle_{flavour}' for flavour in [MEAN, MED, VAR]] \
            + [f'{BRLEN}_tip_middle_{flavour}' for flavour in[MEAN, MED, VAR]] \
            + [f'{BRLEN}_inode_bottom_{flavour}' for flavour in [MEAN, MED, VAR]] \
            + [f'{BRLEN}_tip_bottom_{flavour}' for flavour in[MEAN, MED, VAR]] \
            + [f'frac_{BRLEN}_inode_{flavour}_by_{BRLEN}_tip_{flavour}' for flavour in [MEAN, MED, VAR]] \
            + [f'frac_{BRLEN}_inode_top_{flavour}_by_{BRLEN}_tip_top_{flavour}' for flavour in [MEAN, MED, VAR]] \
            + [f'frac_{BRLEN}_inode_middle_{flavour}_by_{BRLEN}_tip_middle_{flavour}' for flavour in [MEAN, MED, VAR]] \
            + [f'frac_{BRLEN}_inode_bottom_{flavour}_by_{BRLEN}_tip_bottom_{flavour}' for flavour in [MEAN, MED, VAR]]


    def set_forest(self, forest, **kwargs):
        self._forest = forest
        annotate_forest_with_time(self._forest)
        self._idists = None
        self._edists = None
        self._edists_top = None
        self._idists_top = None
        self._edists_middle = None
        self._idists_middle = None
        self._edists_bottom = None
        self._idists_bottom = None
        self._max_time = None


    @property
    def max_time(self):
        if self._max_time is None:
            self._max_time = max(getattr(_, TIME) for tree in self._forest for _ in tree)
        return self._max_time

    @property
    def idists(self):
        if self._idists is None:
            self._idists, self._edists = self._get_branch_lengths()
        return self._idists

    @property
    def edists(self):
        if self._edists is None:
            self._idists, self._edists = self._get_branch_lengths()
        return self._edists

    @property
    def idists_top(self):
        if self._idists_top is None:

            def time_filter(nd):
                return getattr(nd, TIME) <= self.max_time / 3

            self._idists_top, self._edists_top = self._get_branch_lengths(time_filter)
        return self._idists_top

    @property
    def edists_top(self):
        if self._edists_top is None:

            def time_filter(nd):
                return getattr(nd, TIME) <= self.max_time / 3

            self._idists_top, self._edists_top = self._get_branch_lengths(time_filter)
        return self._edists_top

    @property
    def idists_middle(self):
        if self._idists_middle is None:

            def time_filter(nd):
                return (getattr(nd, TIME) <= 2 * self.max_time / 3) \
                    and (self.max_time / 3 < (getattr(nd, TIME) - nd.dist))

            self._idists_middle, self._edists_middle = self._get_branch_lengths(time_filter)
        return self._idists_middle

    @property
    def edists_middle(self):
        if self._edists_middle is None:

            def time_filter(nd):
                return (getattr(nd, TIME) <= 2 * self.max_time / 3) \
                    and (self.max_time / 3 < (getattr(nd, TIME) - nd.dist))

            self._idists_middle, self._edists_middle = self._get_branch_lengths(time_filter)
        return self._edists_middle

    @property
    def idists_bottom(self):
        if self._idists_bottom is None:

            def time_filter(nd):
                return 2 * self.max_time / 3 < (getattr(nd, TIME) - nd.dist)

            self._idists_bottom, self._edists_bottom = self._get_branch_lengths(time_filter)
        return self._idists_bottom

    @property
    def edists_bottom(self):
        if self._edists is None:

            def time_filter(nd):
                return 2 * self.max_time / 3 < (getattr(nd, TIME) - nd.dist)

            self._idists_bottom, self._edists_bottom = self._get_branch_lengths(time_filter)
        return self._edists_bottom

    def _get_branch_lengths(self, node_filter=lambda nd: True):
        internal_dists, external_dists = [], []

        for tree in self._forest:
            for n in tree.traverse():
                if not node_filter(n):
                    continue
                if n.is_leaf():
                    external_dists.append(n.dist)
                elif not n.is_root():
                    internal_dists.append(n.dist)
        return internal_dists, external_dists


    def calculate(self, feature_name, **kwargs):
        if feature_name.startswith(BRLEN):
            is_internal = 'inode' in feature_name
            flavour = feature_name.split('_')[-1]
            if is_internal:
                data = self.idists_top if 'top' in feature_name \
                    else (self.idists_middle if 'middle' in feature_name \
                              else (self.idists_bottom if 'bottom' in feature_name else self.idists))
            else:
                data = self.edists_top if 'top' in feature_name \
                    else (self.edists_middle if 'middle' in feature_name \
                              else (self.edists_bottom if 'bottom' in feature_name else self.edists))
            return BASIC_METRIC2FUN[flavour](data) if len(data) else 0
        if feature_name.startswith('frac'):
            feature_name_1, feature_name_2 = feature_name[len('frac_'): ].split('_by_')
            val1, val2 = self.calculate(feature_name_1), self.calculate(feature_name_2)
            return val1 / max(val2, EPSILON)
        return None

    def help(self, feature_name, *args, **kwargs):
        if feature_name.startswith(BRLEN):
            is_internal = 'inode' in feature_name
            flavour = feature_name.split('_')[-1]
            res = f'{BASIC_METRIC2HELP[flavour]} ' \
                  f'of the lengths of the forest {'non-root internal nodes' if is_internal else 'tips'}.'
            part = 'top' if 'top' in feature_name \
                else ('middle' if 'middle' in feature_name else ('bottom' if 'bottom' in feature_name else ''))
            if part:
                res += f' Only branches starting and finishing in the {part} third of the forest are considered. ' \
                       f'(The total time between the earliest forest tree start and the last sampled tip '\
                       'is split into three equal parts.)'
            return res
        if feature_name.startswith('frac'):
            feature_name_1, feature_name_2 = feature_name[len('frac_'):].split('_by_')
            return f'fraction of two values. The dividend is {self.help(feature_name_1).split('.')[0]}. '\
                   f'The divisor is {self.help(feature_name_1)}'
        return None


import math

import numpy as np
import scipy
from scipy.stats import linregress

from treesumstats.tree_manager import TIME, annotate_forest_with_time


TARGET_AVG_BL = 1

CHERRY_PERCENTILES = [0, 1, 5, 10, 25, 50, 75, 100]
INCUBATION_PERIOD_PERCENTILES = [0, 5, 10, 25, 50, 75, 90, 95, 100]




def get_cherry_metrics(forest):
    """
    Computes the following metrics on tree cherries:
    0th (min), 1st, 5th, 10th, 25th, 50th, 75th and 100th (max) percentiles of cherry tip differences
    0th (min), 1st, 5th, 10th, 25th, 50th, 75th and 100th (max) percentiles of cherry tip differences in reshuffled cherries
    mean and var of tip branch len differences
    mean and var of tip branch len differences in reshuffled cherries
    cherry test value

    :param forest: list(ete3.Tree), forest of trees on which these metrics are computed
    :return: tuple of the above metrics in the above order
    """
    all_cherries = []
    for tree in forest:
        all_cherries.extend(pick_cherries(tree, include_polytomies=True))
    all_cherries = sorted(all_cherries, key=lambda _: getattr(_.root, TIME))

    if all_cherries:
        num_cherries = len(all_cherries)
        random_diffs, real_diffs = get_real_vs_reshuffled_diffs(all_cherries)
        pval = scipy.stats.binomtest((random_diffs < real_diffs).sum(), n=num_cherries, p=0.5, alternative='less').pvalue

        return (*np.percentile(real_diffs, q=CHERRY_PERCENTILES),
                *np.percentile(random_diffs, q=CHERRY_PERCENTILES),
                np.mean(real_diffs), np.var(real_diffs),
                np.mean(random_diffs), np.var(random_diffs), pval)
    else:
        return *np.zeros(len(CHERRY_PERCENTILES)), *np.zeros(len(CHERRY_PERCENTILES)), 0, 0, 0, 0, 1


def get_incubation_metrics(forest):
    """
    Computes the following metrics on tree internal sibling branches:
    percentage of internal nodes that have an internal sibling among all internal nodes
    0th (min), 5th, 10th, 25th, 50th, 75th, 90th, 95th and 100th (max) percentiles of sibling branch len differences
    mean and var of sibling branch len differences
    sibling test value

    :param forest: list(ete3.Tree), forest of trees on which these metrics are computed
    :return: tuple of the above metrics in the above order
    """
    all_siblings = []
    for tree in forest:
        all_siblings.extend(pick_siblings(tree, include_polytomies=True))
    all_siblings = sorted(all_siblings, key=lambda _: getattr(_.root, TIME))

    if all_siblings:
        n_siblings = len(all_siblings)
        random_diffs, real_diffs = get_real_vs_reshuffled_diffs(all_siblings)
        pval = (scipy.stats.binomtest((random_diffs < real_diffs).sum(), n=n_siblings, p=0.5, alternative='greater')
                .pvalue)

        nodes_in_siblings = sum(len(_) for _ in all_siblings)
        num_internal_nodes = sum(sum(1 for _ in tree.traverse() if not _.is_leaf()) for tree in forest)

        return (nodes_in_siblings / num_internal_nodes,
                *np.percentile(real_diffs, q=INCUBATION_PERIOD_PERCENTILES),
                np.mean(real_diffs), np.var(real_diffs), pval)
    else:
        return 0, *np.zeros(len(INCUBATION_PERIOD_PERCENTILES)), 0, 0, 1

def calc_summary_statistics(forest):
    """Calculates the summary statistics for a given forest of trees.

    :param forest: list(ete3.Tree), forest of trees on which statistics are computed

    :return: dict of summary statistics
    """
    annotate_forest_with_time(forest)

    # compute summary statistics based on branch lengths
    summaries = {}



    # compute summary statistics based on cherries:
    summaries.update(zip([f'perc{i}_cdiff' for i in CHERRY_PERCENTILES]
                         + [f'perc{i}_cdiff_random' for i in CHERRY_PERCENTILES]
                         + ['mean_cdiff', 'var_cdiff', 'mean_cdiff_random', 'var_cdiff_random', 'ctest'],
                         get_cherry_metrics(forest)))

    # compute summary statistics based on incubation period:
    summaries.update(zip(['IS_nodes']
                         + [f'perc{i}_sdiff' for i in INCUBATION_PERIOD_PERCENTILES]
                         + ['mean_sdiff', 'var_sdiff', 'stest'],
                         get_incubation_metrics(forest)))



    return summaries







import os

from setuptools import setup, find_packages

setup(
    name='treesumstats',
    packages=find_packages(),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    include_package_data=True,
    package_data={'treesumstats': [os.path.join('..', 'README.md'),
                                   os.path.join('..', 'LICENCE')]},
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    version='0.1',
    description='Encoding phylogenetic trees with summary statistics.',
    author='Anna Zhukova',
    author_email='anna.zhukova@pasteur.fr',
    url='https://github.com/evolbioinfo/treesumstats',
    keywords=['phylogenetics', 'summary statistics', 'phylogenetic trees', 'phylogeny'],
    install_requires=['ete3', 'numpy==2.0.2', "scipy==1.14.1", 'biopython', 'pandas==2.2.3'],
    entry_points={
            'console_scripts': [
                'treesumstats_encode = treesumstats.calculator:main',
            ]
    },
)



import re
from collections import namedtuple

import numpy as np

from treesumstats import FeatureCalculator, NORMALIZED, EPSILON
from treesumstats.tree_manager import annotate_forest_with_time, TIME

TREE_PARTS = ['top', 'middle', 'bottom']

N_LTT_COORDINATES = 20

LTT = namedtuple('LTT', ['times', 'lineages', 'trees'])

def get_ltt_plot(forest):
    """
    Returns a mapping of forest event times to
        the numbers of started-but-unfinished lineages,
        the numbers of started-but-unfinished trees,
    where the times correspond to events (internal or external nodes on the tree)
    and the tree start times (time at the beginning of the root branch).

    :param forest: list(ete3.Tree), forest of trees on which these metrics are computed.
        Its nodes must be annotated with their times (feature treesumstats.tree_manager.TIME)
    :return: named tuple LTT containing three numpy arrays: times, lineages, and trees
    """
    time_dlineages_dtrees = []

    for tree in forest:
        time_dlineages_dtrees.append((getattr(tree, TIME) - tree.dist, 1, 1))
        max_time = -np.inf
        for node in tree.traverse():
            # if is a tip it will give -1, if a binary internal node +1, etc.
            node_time = getattr(node, TIME)
            max_time = max(max_time, node_time)
            time_dlineages_dtrees.append((node_time, len(node.children) - 1, 0))
        time_dlineages_dtrees.append((max_time, 0, -1))


    times, lineages, trees = [], [], []

    cur_lineages, cur_trees = 0, 0
    for time, delta_lineages, delta_trees in sorted(time_dlineages_dtrees, key=lambda _: _[0]):
        cur_lineages += delta_lineages
        cur_trees += delta_trees
        # if there is no collision in times of two events, increase the array
        if not times or np.abs(time - times[-1]) > EPSILON:
            times.append(time)
            lineages.append(cur_lineages)
            trees.append(cur_trees)
        # else just update the last lineage count
        else:
            lineages[-1] = cur_lineages
            trees[-1] = cur_trees

    return LTT(times=np.array(times), lineages=np.array(lineages), trees=np.array(trees))


def normalize_lineages_by_total_tip_count(lineages, total_tip_counts):
    """Normalize lineages by total number of tips in the trees that correspond to each time point"""
    lineages = np.array(lineages, dtype=float)
    mask = total_tip_counts > 0
    lineages[mask] /= total_tip_counts[mask]
    return lineages


def get_ltt_coordinates(ltt, n_coordinates=N_LTT_COORDINATES):
    """
    Returns representation of LTT plot over 'n_coordinates' time intervals,
    distributed equally between the time of the first tree start and the time of the last sampled tip.
    We report the numbers of lineages/trees at the start of each interval,
    and do not end the last interval as it corresponds to the end of the sampling period and zero lineages/trees.

    :param ltt: input LTT plot (named tuple LTT)
    :param n_coordinates: number of coordinates to consider
    :return: output LTT plot (named tuple LTT) over 'n_coordinates' time intervals
    """
    start_time = ltt.times[0]
    end_time = ltt.times[-1]
    dt = (end_time - start_time) / n_coordinates

    # to make absolutely sure its size is n_coordinates + 1 and not (n_coordinates + 2) due to rounding we do [: _]
    xs = np.arange(start_time, end_time + dt, dt)[: (n_coordinates + 1)]
    ys = np.zeros(n_coordinates + 1, dtype=float)
    zs = np.zeros(n_coordinates + 1, dtype=float)
    prev_idx = 0
    for i in range(n_coordinates + 1):
        while prev_idx < len(ltt.times) - 1 and ltt.times[prev_idx + 1] <= xs[i]:
            prev_idx += 1
        ys[i] = ltt.lineages[prev_idx]
        zs[i] = ltt.trees[prev_idx]
    return LTT(times=xs[:-1], lineages=ys[:-1], trees=zs[:-1])


def filter_LTT_by_time(ltt, start_time=-np.inf, stop_time=np.inf):
    mask = (ltt.times >= start_time) & (ltt.times <= stop_time)
    filtered_times = ltt.times[mask]
    filtered_lineages = ltt.lineages[mask]
    filtered_trees = ltt.trees[mask]
    if not len(filtered_times) \
            or (filtered_times[0] > ltt.times[0] and start_time > ltt.times[0] and start_time not in ltt.times):
        prev_i = next(i for i in range(len(ltt.times)) if ltt.times[i + 1] > start_time)
        filtered_times = np.concat([[start_time], filtered_times])
        filtered_trees = np.concat([[ltt.trees[prev_i]], filtered_trees])
        filtered_lineages = np.concat([[ltt.lineages[prev_i]], filtered_lineages])
    if not len(filtered_times) \
            or (filtered_times[-1] < ltt.times[-1] and stop_time < ltt.times[-1] and stop_time not in ltt.times):
        filtered_times = np.concat([filtered_times, [stop_time]])
        filtered_trees = np.concat([filtered_trees, [filtered_trees[-1]]])
        filtered_lineages = np.concat([filtered_lineages, [filtered_lineages[-1]]])
    return LTT(times=filtered_times, lineages=filtered_lineages, trees=filtered_trees)


class LTTFeatureCalculator(FeatureCalculator):
    """Computes tree LTT-related summary statistics."""

    def __init__(self, n_coordinates=N_LTT_COORDINATES):
        self._n_coordinates = n_coordinates
        self._n_tips = None
        self._i_ml_top = None
        self._i_ml_middle = None
        self._i_ml_bottom = None
        self._max_time = None
        self._i_ml = None
        self._ltt = None
        self._forest = None
        self._ltt_bottom = None
        self._ltt_top = None
        self._ltt_middle = None
        self._ltt_coord = None

    def feature_names(self):
        return [f'ltt_time{i}' for i in range(self._n_coordinates)] \
            + [f'ltt_trees{i}' for i in range(self._n_coordinates)] \
            + [f'ltt_lineages{i}' for i in range(self._n_coordinates)] \
            + [f'ltt_lineages{i}_{NORMALIZED}' for i in range(self._n_coordinates)] \
            + ['time_trees_max'] + [f'time_trees_max_{part}' for part in TREE_PARTS] \
            + ['trees_max'] + [f'trees_max_{part}' for part in TREE_PARTS] \
            + ['time_lineages_max'] + [f'time_lineages_max_{part}' for part in TREE_PARTS] \
            + ['lineages_max'] + [f'lineages_max_{part}' for part in TREE_PARTS] \
            + [f'lineages_max_{NORMALIZED}'] + [f'lineages_max_{part}_{NORMALIZED}' for part in TREE_PARTS] \
            + ['lineage_start_to_max_slope', 'lineage_stop_to_max_slope', 'lineage_slope_ratio'] \
            + [f'{f}_{part}' for part in TREE_PARTS for f in ['lineage_start_to_max_slope', 'lineage_stop_to_max_slope', 'lineage_slope_ratio']] \
            + [f'lineage_start_to_max_slope_{NORMALIZED}', f'lineage_stop_to_max_slope_{NORMALIZED}'] \
            + [f'{f}_{part}_{NORMALIZED}' for part in TREE_PARTS for f in ['lineage_start_to_max_slope', 'lineage_stop_to_max_slope']]

    def set_forest(self, forest, **kwargs):
        self._forest = forest
        annotate_forest_with_time(self._forest)
        self._i_ml = None
        self._max_time = None
        self._ltt = None
        self._ltt_bottom = None
        self._ltt_top = None
        self._ltt_middle = None
        self._i_ml_top = None
        self._i_ml_middle = None
        self._i_ml_bottom = None
        self._n_tips = None
        self._ltt_coord = None

    @property
    def ltt(self):
        if self._ltt is None:
            self._ltt = get_ltt_plot(self._forest)
        return self._ltt

    @property
    def ltt_coord(self):
        if self._ltt_coord is None:
            self._ltt_coord = get_ltt_coordinates(self.ltt, n_coordinates=self._n_coordinates)
        return self._ltt_coord

    @property
    def ltt_top(self):
        if self._ltt_top is None:
            self._ltt_top = filter_LTT_by_time(self.ltt, stop_time=self.max_time / 3)
        return self._ltt_top

    @property
    def ltt_middle(self):
        if self._ltt_middle is None:
            self._ltt_middle = filter_LTT_by_time(self.ltt, start_time=self.max_time / 3, stop_time=2 * self.max_time / 3)
        return self._ltt_middle

    @property
    def ltt_bottom(self):
        if self._ltt_bottom is None:
            self._ltt_bottom = filter_LTT_by_time(self.ltt, start_time=2 * self.max_time / 3)
        return self._ltt_bottom

    @property
    def i_ml(self):
        if self._i_ml is None:
            self._i_ml = self.ltt.lineages.argmax()
        return self._i_ml

    @property
    def i_ml_top(self):
        if self._i_ml_top is None:
            self._i_ml_top = self.ltt_top.lineages.argmax()
        return self._i_ml_top

    @property
    def i_ml_middle(self):
        if self._i_ml_middle is None:
            self._i_ml_middle = self.ltt_middle.lineages.argmax()
        return self._i_ml_middle

    @property
    def i_ml_bottom(self):
        if self._i_ml_bottom is None:
            self._i_ml_bottom = self.ltt_bottom.lineages.argmax()
        return self._i_ml_bottom

    @property
    def max_time(self):
        if self._max_time is None:
            self._max_time = max(getattr(_, TIME) for tree in self._forest for _ in tree)
        return self._max_time


    @property
    def n_tips(self):
        if self._n_tips is None:
            self._n_tips = sum(len(tree) for tree in self._forest)
        return self._n_tips


    def calculate(self, feature_name, **kwargs):

        if feature_name.startswith('ltt'):
            coordinate = int(re.findall(r'\d+', feature_name)[0])
            if 'time' in feature_name:
                # Invert the time so that the end of the sampling period is at time 0
                return self.max_time - self.ltt_coord.times[coordinate]
            if 'trees' in feature_name:
                return self.ltt_coord.trees[coordinate]
            if 'lineages' in feature_name:
                denominator = self.n_tips if feature_name.endswith(NORMALIZED) else 1
                return self.ltt_coord.lineages[coordinate] / denominator
            return None

        part = 'top' if '_top' in feature_name \
            else ('middle' if '_middle' in feature_name else ('bottom' if '_bottom' in feature_name else ''))
        if part:
            feature_name = feature_name.replace(f'_{part}', '')
        ltt = self.ltt_top if 'top' == part \
                    else (self.ltt_middle if 'middle' == part \
                              else (self.ltt_bottom if 'bottom' == part else self.ltt))

        if 'time_trees_max' == feature_name:
            # Invert the time so that the end of the sampling period is at time 0
            return self.max_time - ltt.times[ltt.trees.argmax()]
        if 'trees_max' == feature_name:
            return ltt.trees.max()

        i_ml = self.i_ml_top if 'top' == part \
                    else (self.i_ml_middle if 'middle' == part \
                              else (self.i_ml_bottom if 'bottom' == part else self.i_ml))
        if 'time_lineages_max' == feature_name:
            # Invert the time so that the end of the sampling period is at time 0
            return self.max_time - ltt.times[i_ml]

        if 'lineage_slope_ratio' == feature_name:
            time_lin_max, lin_max = ltt.times[i_ml], ltt.lineages[i_ml]
            time_start, lin_start = ltt.times[0], ltt.lineages[0]
            time_stop, lin_stop = ltt.times[-1], ltt.lineages[-1]
            if time_lin_max == time_start == time_stop:
                return 1
            return (lin_max - lin_start) * (time_stop - time_lin_max) \
                / max((time_lin_max - time_start) * (lin_max - lin_stop), EPSILON)

        denominator = self.n_tips if feature_name.endswith(NORMALIZED) else 1
        if feature_name.startswith('lineages_max'):
            return ltt.lineages[i_ml] / denominator

        time_lin_max, lin_max = ltt.times[i_ml], ltt.lineages[i_ml]
        if feature_name.startswith('lineage_start_to_max_slope'):
            time_start, lin_start = ltt.times[0], ltt.lineages[0]
            return (lin_max - lin_start) / max(time_lin_max - time_start, EPSILON) / denominator
        if feature_name.startswith('lineage_stop_to_max_slope'):
            time_stop, lin_stop = ltt.times[-1], ltt.lineages[-1]
            return (lin_max - lin_stop) / max(time_stop - time_lin_max, EPSILON) / denominator

        return None




    def help(self, feature_name, *args, **kwargs):
        if feature_name.startswith('ltt'):
            coordinate = int(re.findall(r'\d+', feature_name)[0])
            explanation = f", where times t_i (0 ≤ i ≤ {self._n_coordinates}) are equally distributed between " \
                          f"the time of the start of the first tree (t_0 = 0) in the forest and " \
                          f"the time of the forest’s last sampled tip (t_{self._n_coordinates}=T)."
            if 'time' in feature_name:
                # Invert the time so that the end of the sampling period is at time 0
                return f"T - t_{coordinate}{explanation}"
            if 'trees' in feature_name:
                return f"number of trees present at t_{coordinate}{explanation}"
            if 'lineages' in feature_name:
                return f'number of lineages present at t_{coordinate}' \
                       f'{' divided by the total number of tips in the forest' if feature_name.endswith(NORMALIZED) else ''}' \
                       f'{explanation}'
            return None

        part = 'top' if '_top' in feature_name \
            else ('middle' if '_middle' in feature_name else ('bottom' if '_bottom' in feature_name else ''))
        explanation = ''
        if part:
            feature_name = feature_name.replace(f'_{part}', '')
            start_time = "the first tree start" if "top" == part else f"{'2' if 'bottom' == part else ''}T/3"
            end_time = "T" if "bottom" == part else f"{'2' if 'middle' == part else ''}T/3"
            explanation = f" in the {part} part of the forest: between {start_time} and {end_time}"
        else:
            start_time = "first tree start"
            end_time = "T"

        if 'time_trees_max' == feature_name:
            return f"T - t_tmax, where T is the sampling time of the last forest's tip " \
                   f"and t_tmax is the time when the maximum number of trees was first achieved{explanation}."
        if 'trees_max' == feature_name:
            return f"maximum number of trees{explanation}."

        if 'time_lineages_max' == feature_name:
            return f"T - t_lmax, where T is the sampling time of the last forest's tip " \
                   f"and t_lmax is the time when the maximum number of lineages was first achieved{explanation}."
        if feature_name.startswith('lineages_max'):
            return f"maximum number of lineages" \
                   f"{explanation}" \
                   f"{', divided by the total number of tips in the forest' if feature_name.endswith(NORMALIZED) else ''}."
        long_exp = f", where t_lmax is the time when the maximum number of lineages was first achieved{explanation}, " \
                   "where T is the sampling time of the last forest's tip."
        if feature_name.startswith('lineage_start_to_max_slope'):
            res = f"linear slope between the lineages at the {start_time} and t_lmax{long_exp}"
            if feature_name.endswith(NORMALIZED):
                res += " This slope is divided by the total number of tips in the forest."
            return res
        if feature_name.startswith('lineage_stop_to_max_slope'):
            res = f"linear slope between the lineages at the t_lmax and {end_time}{long_exp}"
            if feature_name.endswith(NORMALIZED):
                res += " This slope is divided by the total number of tips in the forest."
            return res
        if 'lineage_slope_ratio' == feature_name:
            return f"ratio between the linear slope between the lineages at the {start_time} and t_lmax, " \
                   f"and the linear slope between the lineages at the t_lmax and {end_time}{long_exp}"
        return None


from pandas.core.computation.ops import isnumeric

from treesumstats.basic_sumstats import BasicFeatureCalculator
from treesumstats.branch_sumstats import BranchFeatureCalculator
from treesumstats.event_time_sumstats import EventTimeFeatureCalculator
from treesumstats.ltt_sumstats import LTTFeatureCalculator
from treesumstats.subtree_sumstats import SubtreeFeatureCalculator
from treesumstats.topology_sumstats import TopologyFeatureCalculator
from treesumstats import FeatureRegistry, FeatureManager
from treesumstats.resolution_sumstats import ResolutionFeatureCalculator
from treesumstats.transmission_chain_sumstats import TransmissionChainFeatureCalculator
from treesumstats.tree_manager import read_forest

FeatureRegistry.register(ResolutionFeatureCalculator())
FeatureRegistry.register(TopologyFeatureCalculator())
FeatureRegistry.register(EventTimeFeatureCalculator())
FeatureRegistry.register(BasicFeatureCalculator())
FeatureRegistry.register(BranchFeatureCalculator())
FeatureRegistry.register(TransmissionChainFeatureCalculator())
FeatureRegistry.register(LTTFeatureCalculator())
FeatureRegistry.register(SubtreeFeatureCalculator())



def main():
    """
    Entry point for tree encoding with command-line arguments.
    :return: void
    """
    import argparse

    parser = argparse.ArgumentParser(description="Calculate summary statistics for an input tree/forest.")
    parser.add_argument('--nwk', type=str, required=True, help="input tree/forest files in newick format")
    parser.add_argument('--tab', type=str, required=True, help="output summary statistic table, in tab-delimited format: the first column will contain summary statistic name, and the second its value.")
    parser.add_argument('--statistics', type=str, nargs='*', help='an optional list of summary statistics that should be calculated. If not given all the available statistics will be calculated.')
    parser.add_argument('--add_descriptions', action='store_true', help='whether to add a column "description" with the description of each calculated statistic.')
    params = parser.parse_args()

    forest = read_forest(params.nwk)

    features = params.statistics if params.statistics else FeatureManager.available_features()

    def format_value(value):
        try:
            value = float(value)
            if int(value) == value:
                return f'{value:.0f}'
            if value > 1e-6:
                return f'{value:.6f}'
            return f'{value:.g}'
        except:
            return value


    with open(params.tab, 'w+') as f:
        f.write(f'statistic\tvalue{'\tdescription' if params.add_definitions else ''}\n')
        for feature_name, value in zip(features, FeatureManager.compute_features(forest)):
            f.write(f'{feature_name}\t{format_value(value)}{f"\t{FeatureManager.help(feature_name)}" if params.add_definitions else ""}\n')


if '__main__' == __name__:
    main()


import re

import numpy as np

from treesumstats import BASIC_METRIC2HELP, BASIC_ARRAY_METRICS, BASIC_METRIC2FUN, FeatureCalculator, NORMALIZED

CHAIN_LEN = 4

CHAIN = 'chain'

CHAIN_PERCENTILES = [p for p in range(10, 91, 10) if 50 != p]


class TransmissionChainFeatureCalculator(FeatureCalculator):
    """Computes tree transmission chain-related summary statistics."""

    def __init__(self, chain_len=CHAIN_LEN, percentiles=CHAIN_PERCENTILES):
        self._n_inodes = None
        self._chains = None
        self._chain_len = chain_len
        self._percentiles = percentiles
        self._forest = None

    def feature_names(self):
        return [f'n_{self._chain_len}-{CHAIN}', f'n_{self._chain_len}-{CHAIN}_{NORMALIZED}'] \
            + [f'brlen_sum_{self._chain_len}-{CHAIN}_{flavour}' for flavour in BASIC_ARRAY_METRICS]\
            + [f'brlen_sum_{self._chain_len}-{CHAIN}_perc{p}' for p in self._percentiles]

    def set_forest(self, forest, **kwargs):
        self._forest = forest
        self._chains = None
        self._n_inodes = None

    @property
    def chains(self):
        if self._chains is None:
            self._chains = []

            for tree in self._forest:
                for n in tree.traverse('postorder'):
                    for i in range(1, self._chain_len + 1):
                        feature = f'chain{i}'
                        if i == 1:
                            cur_chain = min([c.dist for c in n.children], default=np.inf)
                        else:
                            prev_feature = f'chain{i - 1}'
                            cur_chain = min([c.dist + getattr(c, prev_feature) for c in n.children], default=np.inf)
                            for c in n.children:
                                delattr(c, prev_feature)

                        if i == self._chain_len:
                            if cur_chain < np.inf:
                                self._chains.append(cur_chain)
                        elif not n.is_root():
                            n.add_feature(feature, cur_chain)

            self._chains = np.array(self._chains)
        return self._chains

    @property
    def n_inodes(self):
        if self._n_inodes is None:
            self._n_inodes = sum(1 for tree in self._forest for n in tree.traverse() if not n.is_leaf())
        return self._n_inodes

    def calculate(self, feature_name, **kwargs):
        if feature_name.startswith(f'n_{self._chain_len}-{CHAIN}'):
            denominator = self.n_inodes if feature_name.endswith(NORMALIZED) else 1
            return len(self.chains) / denominator
        if feature_name.startswith(f'brlen_sum_{self._chain_len}-{CHAIN}'):
            if 0 == len(self.chains):
                return 0
            flavour = feature_name.split('_')[-1]
            if flavour in BASIC_METRIC2FUN:
                return BASIC_METRIC2FUN[flavour](self.chains)
            elif 'perc' in flavour:
                perc = int(re.findall(r'\d+', flavour)[0])
                return np.percentile(self.chains, perc)
        return None

    def help(self, feature_name, *args, **kwargs):
        if feature_name.startswith(f'n_{self._chain_len}-{CHAIN}'):
            return f'number of {self._chain_len}-node transmission chains'\
                   f'{', divided by the number of internal nodes' if NORMALIZED in feature_name else ''}. ' \
                   f'(A {self._chain_len}-node transmission chain of node i is the shortest {self._chain_len}-branch path descending from i [Voznica et al. 2022].)'
        if feature_name.startswith(f'brlen_sum_{self._chain_len}-{CHAIN}'):
            flavour = feature_name.split('_')[-1]
            if flavour in BASIC_METRIC2FUN:
                return f'{BASIC_METRIC2HELP[flavour]} of branch length sums of all {self._chain_len}-node transmission chains. ' \
                   f'(A {self._chain_len}-node transmission chain of node i is the shortest {self._chain_len}-branch path descending from i [Voznica et al. 2022].)'
            elif 'perc' in flavour:
                perc = re.findall(r'\d+', flavour)[0]
                return f'{perc}th percentile of branch length sums of all {self._chain_len}-node transmission chains. ' \
                       f'(A {self._chain_len}-node transmission chain of node i is the shortest {self._chain_len}-branch path descending from i [Voznica et al. 2022].)'
        return None


import os
import re
from collections import Counter
from datetime import datetime

import numpy as np
from Bio import Phylo
from ete3 import Tree, TreeNode

from treesumstats import EPSILON

OLDEST_TIP_TIME = 'oldest_tip_time'

TIME = 'time'

DATE_REGEX = r'[+-]*[\d]+[.\d]*(?:[e][+-][\d]+){0,1}'



def read_nexus(tree_path):
    with open(tree_path, 'r') as f:
        nexus = f.read()
    # replace CI_date="2019(2018,2020)" with CI_date="2018 2020"
    nexus = re.sub(r'CI_date="({})\(({}),({})\)"'.format(DATE_REGEX, DATE_REGEX, DATE_REGEX), r'CI_date="\2 \3"',
                   nexus)
    temp = tree_path + '.{}.temp'.format(datetime.timestamp(datetime.now()))
    with open(temp, 'w') as f:
        f.write(nexus)
    trees = list(Phylo.parse(temp, 'nexus'))
    os.remove(temp)
    return trees


def parse_nexus(tree_path):
    trees = []
    for nex_tree in read_nexus(tree_path):
        todo = [(nex_tree.root, None)]
        tree = None
        while todo:
            clade, parent = todo.pop()
            dist = 0
            try:
                dist = float(clade.branch_length)
            except:
                pass
            name = getattr(clade, 'name', None)
            if not name:
                name = getattr(clade, 'confidence', None)
                if not isinstance(name, str):
                    name = None
            node = TreeNode(dist=dist, name=name)
            if parent is None:
                tree = node
            else:
                parent.add_child(node)
            todo.extend((c, node) for c in clade.clades)
        trees.append(tree)
    return trees


def read_tree(tree_path):
    tree = None
    for f in (3, 2, 5, 0, 1, 4, 6, 7, 8, 9):
        try:
            tree = Tree(tree_path, format=f)
            break
        except:
            continue
    if not tree:
        raise ValueError('Could not read the tree {}. Is it a valid newick?'.format(tree_path))
    return tree


def resolve_tree(tree, max_extra_brlen=0):
    """
    Resolves polytomies in the tree in a coalescent manner.
    The newly created branch gets a length uniformly drawn
    from ]0; min(max_extra_brlen, min(99% of coallessed child branch lengths)],
    which is then removed from the corresponding child branch lengths.

    :param tree: tree to resolve
    :param max_extra_brlen: maximum branch length for newly created branches
    :return:
    """
    polytomy_counter = Counter()
    for n in tree.traverse('postorder'):
        n_my = len(n.children)
        if n_my > 2:
            polytomy_counter[n_my] += 1
            while len(n.children) > 2:
                child1, child2 = np.random.choice(n.children, 2, replace=False)
                n.remove_child(child1)
                n.remove_child(child2)
                dist = (1 - np.random.random(1)[0]) * min(max_extra_brlen, child1.dist * .99, child2.dist * .99) \
                    if max_extra_brlen > 0 else 0
                parent = n.add_child(dist=dist)
                parent.add_child(child1, dist=child1.dist - dist)
                parent.add_child(child2, dist=child2.dist - dist)
    print('Resolved {} polytomies in a tree of {} tips: {}'
          .format(sum(polytomy_counter.values()), len(tree),
                  ', '.join('{} of {}'.format(v, k)
                            for (k, v) in sorted(polytomy_counter.items(), key=lambda _: -_[0]))))


def resolve_forest(forest, max_extra_brlen=None):
    """
    Resolves polytomies in the forest in a coalescent manner.
    The newly created branch gets a length uniformly drawn
    from ]0; min(max_extra_brlen, min(99% of coallessed child branch lengths)],
    which is then removed from the corresponding child branch lengths.

    :param forest: forest to resolve
    :param max_extra_brlen: maximum branch length for newly created branches.
    If not given (None), will be set to 1% of the length of the shortest non-zero branch in the tree.
    :return:
    """
    if not max_extra_brlen:
        max_extra_brlen = min(min(_.dist for _ in tree.traverse() if _.dist) for tree in forest) * 0.01

    for tree in forest:
        resolve_tree(tree, max_extra_brlen)


def rescale_forest(forest, T_target=1000, T=None):
    annotate_forest_with_time(forest)
    T_initial = get_T(T=T, forest=forest)

    if T_initial == T_target:
        return 1

    scaling_factor = T_target / T_initial

    for tree in forest:
        for n in tree.traverse():
            n.add_feature(TIME, getattr(n, TIME) * scaling_factor)
            n.dist *= scaling_factor

    return scaling_factor


def name_tree(tre):
    """
    Names all the tree nodes with unique names.
    :param tre: ete3.Tree, the tree to be named
    :return: void, modifies the original tree
    """
    for i, node in enumerate(tre.traverse('levelorder')):
        node.name = i

def read_forest(tree_path):
    try:
        roots = parse_nexus(tree_path)
        if roots:
            return roots
    except:
        pass
    if os.path.exists(tree_path):
        with open(tree_path, 'r') as f:
            nwks = f.read().replace('\n', '').split(';')
    else:
        try:
            nwks = tree_path.replace('\n', '').split(';')
        except:
            pass
    if not nwks:
        raise ValueError('Could not find any trees (in newick or nexus format) in the file {}.'.format(tree_path))
    return [read_tree(nwk + ';') for nwk in nwks[:-1]]


def annotate_tree_with_time(tree, start_time=0):
    for n in tree.traverse('preorder'):
        if not hasattr(n, TIME):
            p_time = start_time if n.is_root() else getattr(n.up, TIME)
            n.add_feature(TIME, p_time + n.dist)
    return tree


def annotate_forest_with_time(forest, start_times=None):
    if start_times:
        if len(start_times) < len(forest):
            raise ValueError(f'{len(start_times)} start times are specified but the forest contains {len(forest)} trees. '
                             f'Either specify as many start times as forest trees or set start times to None '
                             f'(to put all the tree start times at 0)')
    else:
        start_times = [0] * len(forest)

    for tree, start_time in zip(forest, start_times):
        if not hasattr(tree, TIME):
            annotate_tree_with_time(tree, start_time)


def get_T(T, forest):
    if T is None:
        T = 0
        for tree in forest:
            T = max(T, max(getattr(_, TIME) for _ in tree))
    return T


def sort_tree(tree, add_root_feature=False, oldest_tip_time_feature=OLDEST_TIP_TIME):
    """
    Reorganise a tree in such a way that for each node its child subtrees are sorted by the time of sampling:
    the subtree containing the oldest tip (with the oldest sampling time) is the first.
    The tree must be time-annotated.

    :param tree: input tree as a ete3 object, with TIME annotations
    :param add_root_feature: if True, the root will get annotated with the time of its oldest sampled tip
    :param oldest_tip_time_feature: feature name to store the root's oldest tip's sampling time
    :return: modified tree (it modifies the input tree object and returns it)
    """
    for n in tree.traverse('postorder'):
        if n.is_leaf():
            n.add_feature(oldest_tip_time_feature, getattr(n, TIME))
            continue
        n.children = sorted(n.children, key=lambda _: getattr(_, oldest_tip_time_feature))
        min_t = np.inf
        for c in n.children:
            min_t = min(min_t, getattr(c, oldest_tip_time_feature))
            delattr(c, oldest_tip_time_feature)
        if not n.is_root() or add_root_feature:
            n.add_feature(oldest_tip_time_feature, min_t)
    return tree


def sort_forest(forest, oldest_tip_time_feature=OLDEST_TIP_TIME):
    """
    Reorganises each tree in the forest in such a way that for each node
    its child subtrees are sorted by the time of sampling:
    the subtree containing the oldest tip (with the oldest sampling time) is the first.
    Then sorts the forest by the time of sampling:
    the tree containing the oldest tip (with the oldest sampling time) is the first.
    The forest must be time-annotated.

    :param forest: list of input trees as ete3 objects, with TIME annotations
    :param oldest_tip_time_feature: feature name to store the root's oldest tip's sampling time
    :return: list of sorted trees
    """

    for tree in forest:
        sort_tree(tree, add_root_feature=True, oldest_tip_time_feature=oldest_tip_time_feature)
    return sorted(forest, key=lambda _: getattr(_, oldest_tip_time_feature))


def tree2vector(tree, sort=True):
    if sort:
        sort_tree(tree)

    def node2vec(node):
        result = []
        for child in node.children:
            result.extend(node2vec(child))
        result.append((getattr(node, TIME) - node.dist, getattr(node, TIME)))
        return result

    result = node2vec(tree)
    if tree.dist >= EPSILON:
        start_time = getattr(tree, TIME) - tree.dist
        result.append((start_time, start_time))

    return result


def forest2vector(forest):
    forest = sort_forest(forest)
    result = []
    for tree in forest:
        result.extend(tree2vector(tree, sort=False))
    return result


def vector2forest(vector):
    result = []
    while vector:
        result.append(vector2tree(vector))
    return list(reversed(result))


def vector2tree(vector):

    def vec2tree(vec, tree=None):
        if not vec:
            return tree
        tp, ti = vec.pop()
        node = TreeNode(dist=ti - tp)
        node.add_feature(TIME, ti)

        if not tree or ti > tp or np.abs(getattr(tree, TIME) - tp) > EPSILON:
            # A zero-length branch means it is a root of the next tree
            while vec and np.abs(vec[-1][0] - ti) < EPSILON <= np.abs(vec[-1][0] - vec[-1][1]):
                node = vec2tree(vec, node)

        if not tree:
            tree = node
        elif np.abs(getattr(tree, TIME) - tp) < EPSILON:
            tree.add_child(node)
        elif np.abs(getattr(tree, TIME) - tree.dist - ti) < EPSILON:
            node.add_child(tree)
            tree = node

        return tree

    tree = vec2tree(vector)
    for n in tree.traverse('postorder'):
        n.children = list(reversed(n.children))
    if len(tree.children) == 1 and tree.dist < EPSILON:
        tree = tree.children[0].detach()
    return tree

